"""
Conch.py, a music toolkit.
By Kris Schnee, borrowing heavily from Pygame's docs and examples.
 
License: Free software; use as you please. Credit appreciated.
Requirements: Just Python and Pygame. Put all sound and music files
in subdirectories called "sound" and "music".
Notes: These are easy functions for using music and sound in Python/Pygame,
just wrappers around Pygame's functions. The Jukebox class keeps track of a
set of loaded sound effects and paths for songs, with keys so you can just
call 'j.PlaySong("Battle Music")' to access an obscurely named song file
neatly stored in a subdirectory. So, to use this code you just LoadSong for
whatever songs you like, giving the filename and a nickname, then PlaySong
to play. Same for sound effects, though behind the scenes the sounds are
actually loaded once and kept in memory instead of just the paths.
 
Example:
j = Jukebox()
j.LoadSong("battle00.mid","Battle Music")
j.PlaySong("Battle Music")
j.StopMusic()
"""
 
 
import pygame
import os
import constants
MODULE_NAME = "Conch"
MODULE_VERSION = "2006.8.9"
 
DEFAULT_MUSIC_EXTENSION = ".mid"
DEFAULT_SOUND_EXTENSION = ".wav"
SOUND_DIRECTORY = "data"
MUSIC_DIRECTORY = "data"
JUKEBOX_COMMENTS = False
MUSIC_ON = True
SOUND_ON = True
 
class Jukebox:
    def __init__(self,dir=MUSIC_DIRECTORY):
        """Load and play sounds and music, referenced by name.
        Create one of these to put audio in your game.
        One is created automatically when the module loads, so
        there's not really a need to make another."""
        self.name = "Jukebox"
        self.comments = JUKEBOX_COMMENTS
 
        self.music_on = MUSIC_ON
        self.sound_on = SOUND_ON
 
        self.songs = {} ## eg. {"Battle Theme":"battle01.ogg"}
 
        self.music_directory = MUSIC_DIRECTORY
        self.sound_directory = SOUND_DIRECTORY
 
        self.sounds = {}
 
    def Comment(self,what):
        if self.comments:
            print "["+self.name+"] " + str(what)
 
    def ToggleMusic(self,on=True):
        self.music_on = on
 
    def ToggleSound(self,on=True):
        self.sound_on = on
 
    def StopMusic(self):
        """Stops music without turning it off; another song may get cued."""
        pygame.mixer.music.stop()
 
    def QuitMusic(self):
        """Shuts off Pygame's music code.
        This probably isn't necessary."""
        pygame.mixer.music.stop()
        pygame.mixer.quit()
 
    def LoadSong(self,songname,key=""):
        """Add name, including directory location, to songlist.
        You can give the song a key, too, for easy reference.
        Note that rather than actually loading the song, we store only
        the path to it. Contrast with sound loading."""
        if not self.music_on: return
        new_song_path = os.path.join(self.music_directory,songname)
        if key:
            self.songs[ key ] = new_song_path
 
    def ResetSongData(self):
        self.songs = {}
 
    def play_song(self,cue_name, queue=False):
        """Cue this song. If "interrupt," the song will start even
        if one is already playing."""
        if not self.music_on:
            return
        path = self.songs.get( cue_name )

        if path:
            pygame.mixer.music.stop()
 
            try:
                if not queue: pygame.mixer.music.load(path)
                else: pygame.mixer.music.queue(path)
            except:
                print "Couldn't load song '"+cue_name+"'."
                return
            try:
                if not queue: 
                    pygame.mixer.music.play()
                    self.Comment("Cue music: '"+cue_name+"'")
            except:
                print "Couldn't play song '"+cue_name+"'."

    def LoadSound(self,filename,cue_name=None):
        """Load a sound into memory, not just its name, for quick use."""
        if not self.sound_on: return
        if not "." in filename:
            filename += DEFAULT_SOUND_EXTENSION
        try:
            new_sound = pygame.mixer.Sound( os.path.join(SOUND_DIRECTORY,filename) )
            new_sound.set_volume(0.2) #phew. fricking sounds
            if not cue_name:
                cue_name = filename
            self.sounds[ cue_name ] = new_sound
        except Exception, inst:
            print inst.message 
            print "\n > Try the 'nosound' switch if you want to run the game with no sound\n"
            self.turn_off()
 
    def play_sound(self,cue_name):
        if not self.sound_on: return
        ## How to check whether sound player is busy?
        if self.sounds.has_key( cue_name ):
            self.sounds[ cue_name ].play()
        else:
            self.Comment("Tried to play sound '"+cue_name+"' without loading it.")
            pass

    def turn_off(self):
        self.music_on = False
        self.sound_on = False

    def init(self):
        try:
            pygame.mixer.init()
        except:
            print 'could not init sound mixer'
 
        juke.LoadSound("ouch.wav", constants.OUCH_SOUND)
        juke.LoadSound("noo.wav", constants.NOO_SOUND)
        juke.LoadSound("jump.wav", constants.JUMP_SOUND)
        juke.LoadSong("electronic_ghostly.mp3", constants.GHOSTLY_MUSIC)
        juke.LoadSong("hip_hop_synth.mp3", constants.HIP_HOP_MUSIC)
        juke.LoadSong("jazz_rock.mp3", constants.JAZZ_ROCK_MUSIC)
 
juke = Jukebox()
