from World import character, string
import constants
from math import sqrt

class TestCharacter(object):
    def setup_method(self, method):
        self.hero = character.Character()

    def test_should_move_up_when_not_at_top_of_string(self):
        x, y = self.hero.position()
        rope = string.String(x, y-1)
        self.hero.grab_string(rope)
        self.hero.climb_up()
        self.assert_position(x, y-constants.CLIMB_SPEED)

    def test_should_not_move_up_when_at_top_of_string(self):
        x, y = self.hero.position()
        rope = string.String(x, y)
        self.hero.grab_string(rope)
        self.hero.climb_up()
        self.assert_position(x, y)

    def test_should_attach_to_highest_and_lowest_string_met(self):
        lowest = string.String(x=0, y=20, length=100)
        highest = string.String(x=0, y=10, length=100)
        self.hero.x = 0
        self.hero.y = 20
        self.hero.collide(lowest)
        self.hero.collide(highest)
        assert self.hero._string == lowest
        self.hero.attach_to_highest_string_met()
        assert self.hero._string == highest
        self.hero.attach_to_lowest_string_met()
        assert self.hero._string == lowest

    def test_should_move_down_when_not_at_bottom_of_string(self):
        x, y = self.hero.position()
        rope = string.String(x, y-1, 10)
        self.hero.grab_string(rope)
        self.hero.climb_down()
        self.assert_position(x, y+constants.CLIMB_SPEED)
        assert self.hero.is_on_string()

    def test_should_fall_off_when_climb_off_bottom_of_string(self):
        x, y = self.hero.position()
        rope = string.String(x, y, 0)
        self.hero.grab_string(rope)
        self.hero.climb_down()
        self.assert_position(x, y+constants.CLIMB_SPEED)
        assert not self.hero.is_on_string()

    def test_should_follow_curved_path_when_jumping_left(self):
        x, y = self.hero.position()
        self.put_hero_on_string()
        self.hero.jump_left(angle=60, speed=2)
        self.hero.tick_update()
        self.assert_position(x-character.Character.STRAFE_OFFSET-1, y + self.displacement_due_to_acceleration(-sqrt(3), constants.GRAVITY, 1))
        self.hero.tick_update()
        self.assert_position(x-character.Character.STRAFE_OFFSET-2, y + self.displacement_due_to_acceleration(-sqrt(3), constants.GRAVITY, 2))
        self.hero.tick_update()
        self.assert_position(x-character.Character.STRAFE_OFFSET-3, y + self.displacement_due_to_acceleration(-sqrt(3), constants.GRAVITY, 3))
        self.hero.tick_update()
        self.assert_position(x-character.Character.STRAFE_OFFSET-4, y + self.displacement_due_to_acceleration(-sqrt(3), constants.GRAVITY, 4))

    def test_should_follow_curved_path_when_jumping_right(self):
        x, y = self.hero.position()
        self.put_hero_on_string()
        self.hero.jump_right(angle=60, speed=2)
        self.hero.tick_update()
        self.assert_position(x+character.Character.STRAFE_OFFSET+1, y + self.displacement_due_to_acceleration(-sqrt(3), constants.GRAVITY, 1))
        self.hero.tick_update()
        self.assert_position(x+character.Character.STRAFE_OFFSET+2, y + self.displacement_due_to_acceleration(-sqrt(3), constants.GRAVITY, 2))
        self.hero.tick_update()
        self.assert_position(x+character.Character.STRAFE_OFFSET+3, y + self.displacement_due_to_acceleration(-sqrt(3), constants.GRAVITY, 3))
        self.hero.tick_update()
        self.assert_position(x+character.Character.STRAFE_OFFSET+4, y + self.displacement_due_to_acceleration(-sqrt(3), constants.GRAVITY, 4))

    def test_can_attach_to_a_string(self):
        assert not self.hero.is_on_string()
        self.put_hero_on_string()
        assert self.hero.is_on_string()

    def test_can_jump_when_on_string(self):
        original_position = self.hero.position()
        self.put_hero_on_string()

        self.hero.jump_left(angle=45, speed=sqrt(2))
        self.hero.tick_update()
        assert self.hero.is_above(*original_position)

    def test_can_not_jump_when_not_on_string(self):
        original_position = self.hero.position()

        self.hero.jump_left(angle=45, speed=sqrt(2))
        self.hero.tick_update()
        assert not self.hero.is_above(*original_position)

    def test_character_should_be_invincible_for_a_little_while(self):
        lives_before = self.hero.life().remaining()
        self.hero.hit()
        lives_after = self.hero.life().remaining()
        assert lives_before == lives_after + 1
        self.hero.hit()
        assert lives_after == self.hero.life().remaining()
    
    def test_should_not_fall_when_attached(self):
        original_position = self.hero.position()
        self.put_hero_on_string()

        self.hero.tick_update()

        assert not self.hero.is_below(*original_position)

    def test_should_fall_when_not_attached(self):
        original_position = self.hero.position()
        
        self.hero.tick_update()

        assert not self.hero.is_below(*original_position)

    def test_should_no_longer_be_on_string_once_it_has_burned_up(self):
        x, y = self.hero.position()
        rope = string.String(x, y-2, 3)
        self.hero.grab_string(rope)

        assert self.hero.is_on_string()
        rope.burn(rope.length())
        assert not self.hero.is_on_string()

    def test_hero_should_die_when_life_under_zero(self):
        for i in range(constants.FULL_LIFE):
            self.hero._hit()
           
        assert self.hero.is_dead()

    def test_hero_should_not_die_when_life_is_above_zero(self):
        for i in range(constants.FULL_LIFE - 2):
            self.hero.hit()

        assert not self.hero.is_dead()
        
    def test_hero_should_understand_his_height(self):
        self.put_hero_on_string()
        
        original_height = self.hero.height()
        assert 0 == original_height
        
        self.hero.climb_up()
        
        climb_up_height = self.hero.height()
        assert original_height < climb_up_height
        
        self.hero.climb_down()
        self.hero.climb_down()
        climb_down_twice_height = self.hero.height()
        assert climb_down_twice_height < climb_up_height
        
    def test_hero_should_in_repositioning_when_he_is_hitten(self):
        assert not self.hero.is_repositioning()
        self.hero.hit()
        assert self.hero.is_repositioning()
        
    def test_hero_should_reposition_after_he_is_hitten(self):
        self.hero.hit()
        self.hero.reposition()
        assert not self.hero.is_repositioning()
        assert self.hero.is_on_string()
    ###

    def put_hero_on_string(self):
        (x, y) = self.hero.position()
        self.hero.grab_string(string.String(x, y - 8))

    def assert_position(self, expected_x, expected_y):
        actual_x, actual_y = self.hero.position()
        assert round(expected_x, 2) == round(actual_x, 2)
        assert round(expected_y, 2) == round(actual_y, 2)
    
    def displacement_due_to_acceleration(self, u, a, t):
        """Because this is discrete, not continuous, we have to modify the formula slightly"""
        return u*t + 0.5*a*t*(t-1)

